# SWMODD

`SWMODD` is a software module daemon to manage software modules as defined in [TR181-usp](https://usp-data-models.broadband-forum.org/tr-181-2-14-1-usp.html#D.Device:2.Device.SoftwareModules.) and [TR181-cwmp](https://cwmp-data-models.broadband-forum.org/tr-181-2-14-1-cwmp.html#D.Device:2.Device.SoftwareModules.) datamodels.

It is written in C programming language and depends on a number of libraries of OpenWrt for building and running.

## Good to Know
`SWMODD` currently has support for LXC containers and can also be used to manage host services in the same way. To manage the service/deployment units it depends on `opkg`, so all the `opkg` limitations can be considered as limitation for `swmodd` as well with respect to listing/installing/upgrading packages/services.

Currently, `swmodd` can not create new execution environment and it depends on pre-created execution environments/containers. `swmodd` treats host as one execution environment and further it usages the pre-created lxc-containers as available execution environments for managing services.

NOTE:
- `SWMODD` in du_list only shows information about the packages from `opkg list` which has services registered in init.d
- `SWMODD` in eu_list only shows information about the packages present in `ubus call service list`
- `SWMODD` maintains a list of mandatory packages such as procd, ubusd, ubus client, swmodd and opkg as of now. Information about these packages are not shown in both eu_list and du_list.

## Concepts and Workflow
`SWMODD` usages lxc library to interact with the lxc containers and opkg system utilities to manage the services running inside that container.
`swmodd` is used to manage the software modules and exposes the functionality over ubus, whereas `libswmodd.so` is `bbf` plugin which exposes the SoftwareModules functionality over TR181 using `libbbf_api`.

## swmodd uBus

`swmodd` must be launched on startup after `uspd`. It exposes the SoftwareModules datamodel objectes over ubus. The `swmodd` registers `swmodules` namespaces with ubus, that has the shown below functionalities:

```bash
root@iopsys:~# ubus -v list swmodules
'swmodules' @b09f9fc4
        "ee_list":{}
        "du_list":{"eeid":"Integer","ee_name":"String"}
        "eu_list":{"eeid":"Integer","ee_name":"String"}
        "du_install":{"eeid":"Integer","ee_name":"String","uuid":"String","url":"String","username":"String","password":"String"}
        "du_update":{"eeid":"Integer","ee_name":"String","uuid":"String","url":"String","username":"String","password":"String"}
        "du_uninstall":{"eeid":"Integer","ee_name":"String","du_name":"String"}
        "eu_set_state":{"eeid":"Integer","ee_name":"String","eu_name":"String","state":"Boolean"}
        "ee_set_state":{"eeid":"Integer","ee_name":"String","state":"String"}
root@iopsys:~#
```

For more info on the `swmodules` ubus schema see [link](./docs/api/swmodules.md) or [raw schema](./schemas/ubus/swmodules.json)

## swmodd lxc templates
swmodd also includes LXC templates to easily create LXC based containers.

#### To create a new container using iopsys template:

`lxc-create -n <container_name> -t iopsys`
example:
```bash
root@iopsys:~# lxc-create -n test -t iopsys
root@iopsys:~#
```

#### To list containers:
```bash
root@iopsys:~# lxc-ls -f
NAME  STATE   AUTOSTART GROUPS IPV4          IPV6                               UNPRIVILEGED
test  STOPPED 0         -      -             -                                  false
test2 RUNNING 0         -      -             fde4:fbc2:5507:0:cc7d:3ff:febe:ba1 false
test3 RUNNING 0         -      192.168.1.209 -                                  false
root@iopsys:~#
```

#### To start a container:
```
root@iopsys:~# lxc-start -n test
```

#### To execute command inside container
```bash
root@iopsys:/srv/lxc# lxc-attach -n test
BusyBox v1.33.1 (2021-07-28 21:10:07 IST) built-in shell (ash)

root@test:/#
root@test:/# udhcpc
udhcpc: started, v1.33.1
udhcpc: sending discover
udhcpc: sending discover
udhcpc: sending select for 192.168.1.209
udhcpc: lease of 192.168.1.209 obtained, lease time 3600
root@test:/#
root@test:~# ifconfig
eth0      Link encap:Ethernet  HWaddr AE:16:60:98:DD:70
          inet addr:192.168.1.209  Bcast:192.168.1.255  Mask:255.255.255.0
          inet6 addr: fde4:fbc2:5507:0:ac16:60ff:fe98:dd70/64 Scope:Global
          UP BROADCAST RUNNING MULTICAST  MTU:1500  Metric:1
          RX packets:267 errors:0 dropped:15 overruns:0 frame:0
          TX packets:40 errors:0 dropped:0 overruns:0 carrier:0
          collisions:0 txqueuelen:1000
          RX bytes:19556 (19.0 KiB)  TX bytes:3204 (3.1 KiB)

lo        Link encap:Local Loopback
          inet addr:127.0.0.1  Mask:255.0.0.0
          inet6 addr: ::1/128 Scope:Host
          UP LOOPBACK RUNNING  MTU:65536  Metric:1
          RX packets:0 errors:0 dropped:0 overruns:0 frame:0
          TX packets:0 errors:0 dropped:0 overruns:0 carrier:0
          collisions:0 txqueuelen:1000
          RX bytes:0 (0.0 B)  TX bytes:0 (0.0 B)

root@test:~# exit
```

#### To stop container:

```bash
root@iopsys:~# lxc-stop -n test
```

#### To destroy pre-existing container:
```bash
root@iopsys:~# lxc-destroy -n test
root@iopsys:~#
```

### swmodd ubus examples
The outputs shown below are just an example, it can vary on each system. Long outputs are truncated to beautify the document.

#### List down the available execution environments
```bash
root@iopsys:~# ubus call swmodules ee_list
{
        "environment": [
                {
                        "ee_name": "OpenWRT_Linux",
                        "eeid": 1,
                        "status": "Up",
                        "pause": 0,
                        "autoboot": 1,
                        "type": "Linux",
                        "vendor": "iopsys",
                        "version": "4.19.183",
                        "allocated_disk_space": 32568,
                        "available_disk_space": 28672,
                        "allocated_memory": 506904,
                        "available_memory": 313256,
                        "parent_ee_ref": 0
                },
        ]
}
root@iopsys:~#
```
#### List available deployment units/installed services
```bash
root@iopsys:~# ubus call swmodules du_list
{
        "deployment_unit": [
                {
                        "du_name": "at",
                        "ee_name": "OpenWRT_Linux",
                        "eeid": "1",
                        "uuid": "c560bb13-7058-4020-b05c-609ced9a306e",
                        "duid": "vcc0000",
                        "url": "",
                        "version": "3.2.1-1",
                        "config": "",
                        "description": "At and batch read shell commands from standard input storing them as a job to",
                        "vendor": ""
                },
		...
		{
                        "du_name": "xl2tpd",
                        "ee_name": "OpenWRT_Linux",
                        "eeid": "1",
                        "uuid": "ad1b59f6-7557-4b0a-8596-65f8f883872f",
                        "duid": "vcc0049",
                        "url": "",
                        "version": "1.3.16-2",
                        "config": "",
                        "description": "l2tpd is the open source implementation of the L2TP tunneling protocol (RFC2661).",
                        "vendor": ""
                }
        ]
}
```

#### List execution units/running services
```bash
root@iopsys:~# ubus call swmodules eu_list
{
        "execution_unit": [
                {
                        "eu_name": "atd",
                        "command": "/usr/sbin/atd",
                        "state": "Active",
                        "config": "",
                        "version": "",
                        "description": "",
                        "ee_name": "OpenWRT_Linux",
                        "eeid": 1,
                        "euid": 6020,
                        "disk_space": 0,
                        "memory_space": 840,
                        "vendor": ""
                },
		...
		{
                        "eu_name": "xl2tpd",
                        "command": "xl2tpd",
                        "state": "Active",
                        "config": "",
                        "version": "",
                        "description": "",
                        "ee_name": "OpenWRT_Linux",
                        "eeid": 1,
                        "euid": 6492,
                        "disk_space": 0,
                        "memory_space": 856,
                        "vendor": ""
                }
        ]
}
```

#### Install new deployment unit/service
```bash
root@iopsys:~# ubus call swmodules du_install '{"url":"http://download.iopsys.eu/iopsys/packages/iopsys-brcm63xx-arm/6.2.0BETA3/wfadatad_4.0.2_arm_xscale.ipk", "ee_name":"OpenWRT_Linux"}'
{
        "status": true
}
root@iopsys:~#
```
#### Uninstall already installed deployment units/remove services
```bash
root@iopsys:~# ubus call swmodules du_uninstall '{"du_name":"wfadatad", "ee_name":"OpenWRT_Linux"}'
{
        "status": true
}
```
#### Change the state of execution units
```bash
root@iopsys:~# ubus call swmodules eu_activate '{"eeid":1, "eu_name":"obuspa", "state":false}'
{
        "status": true
}
root@iopsys:~#
```
#### Change the state of execution environment
Possible states of execution environment could be `start`, `stop`, `pause` and `resume`. To stop, pause or resume an execution environment the environment must be started prior.

```bash
root@iopsys:~# ubus call swmodules ee_set_state '{"eeid":2, "ee_name":"test", "state":"start"}'
{
        "status": true,
        "reason": ""
}
root@iopsys:~#
root@iopsys:~# ubus call swmodules ee_set_state '{"eeid":2, "ee_name":"test", "state":"stop"}'
{
        "status": true,
        "reason": ""
}
root@iopsys:~#
root@iopsys:~# ubus call swmodules ee_set_state '{"eeid":2, "ee_name":"test", "state":"pause"}'
{
        "status": false,
        "reason": "ExecEnv not running"
}
root@iopsys:~#
root@iopsys:~# ubus call swmodules ee_set_state '{"eeid":3, "ee_name":"test", "state":"resume"}'
{
        "status": false,
        "reason": "no environment exist at this index"
}
root@iopsys:~#
```

## Limitations
- `swmodd` can only manage host and pre-created lxc containers

## Dependencies
`swmodd` compile time and run time dependencies.

### Build-Time Dependencies

To successfully build `swmodd`, the following libraries are needed:

| Dependency  | Link                                          | License                                |
| ----------- | --------------------------------------------- | -------------------------------------- |
| libuci      | https://git.openwrt.org/project/uci.git       | LGPL 2.1                               |
| libubox     | https://git.openwrt.org/project/libubox.git   | BSD                                    |
| ubus        | https://git.openwrt.org/project/ubus.git      | LGPL 2.1                               |
| liblxc      | https://linuxcontainers.org/downloads/lxc/    | LGPL-2.1-or-later BSD-2-Clause GPL-2.0 |
| opkg        | https://git.openwrt.org/project/opkg-lede.git | GPL-2.0                                |
| libuuid     | util-linux                                    | GPL-2.0                                |
| libcurl     | https://dl.uxnr.de/mirror/curl                | MIT                                    |


### Run-Time Dependencies

In order to run the `swmodd`, following dependencies are needed to be running/available before `swmodd`.

| Dependency |                   Link                        | License  |
| ---------- | --------------------------------------------- | -------- |
| ubusd      | https://git.openwrt.org/project/ubus.git      | LGPL 2.1 |
| bbf        | https://dev.iopsys.eu/iopsys/bbf.git          | LGPL 2.1 |
| opkg       | https://git.openwrt.org/project/opkg-lede.git | GPL-2.0  |
| uspd       | https://dev.iopsys.eu/iopsys/uspd.git         | GPL-2.0  |
| libcurl    | https://dl.uxnr.de/mirror/curl                | MIT      |

Apart from these dependencies, LXC-4.0 > also requires cgroup functionalities to be mounted separately, to do so, `cgroupfs-mount` package is used.
> `uspd` along with `bbf` is used to expose the SoftwareModule datamodel objects.

