/*
 * Copyright (C) 2021 iopsys Software Solutions AB
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2.1
 * as published by the Free Software Foundation
 *
 * Author: Amin Ben Ramdhane <amin.benramdhane@pivasoftware.com>
 */

#include "datamodel.h"

/* ********** DynamicObj ********** */
DM_MAP_OBJ tDynamicObj[] = {
/* parentobj, nextobject, parameter */
{"Device.", tSWmodObj, NULL},
{0}
};

/**************************************************************************
* LINKER
***************************************************************************/
static int get_exec_env_linker(char *refparam, struct dmctx *dmctx, void *data, char *instance, char **linker)
{
	*linker = dmjson_get_value((json_object *)data, 1, "ee_name");
	return 0;
}

static int get_du_linker(char *refparam, struct dmctx *dmctx, void *data, char *instance, char **linker)
{
	char *name = dmjson_get_value((json_object *)data, 1, "du_name");
	char *environment = dmjson_get_value((json_object *)data, 1, "ee_name");
	dmasprintf(linker, "%s-%s", name, environment);
	return 0;
}

/*************************************************************
* ENTRY METHOD
*************************************************************/
static int browseSoftwareModulesExecEnvInst(struct dmctx *dmctx, DMNODE *parent_node, void *prev_data, char *prev_instance)
{
	json_object *res = NULL, *du_obj = NULL, *arrobj = NULL;
	char *inst = NULL;
	int id = 0, env = 0;

	dmubus_call("swmodules", "ee_list", UBUS_ARGS{0}, 0, &res);
	dmjson_foreach_obj_in_array(res, arrobj, du_obj, env, 1, "environment") {
		inst = handle_instance_without_section(dmctx, parent_node, ++id);
		if (DM_LINK_INST_OBJ(dmctx, parent_node, (void *)du_obj, inst) == DM_STOP)
			break;
	}
	return 0;
}

static int browseSoftwareModulesDeploymentUnitInst(struct dmctx *dmctx, DMNODE *parent_node, void *prev_data, char *prev_instance)
{
	json_object *res = NULL, *du_obj = NULL, *arrobj = NULL;
	char *inst = NULL;
	int id = 0, du = 0;

	dmubus_call("swmodules", "du_list", UBUS_ARGS{0}, 0, &res);
	dmjson_foreach_obj_in_array(res, arrobj, du_obj, du, 1, "deployment_unit") {
		inst = handle_instance_without_section(dmctx, parent_node, ++id);
		if (DM_LINK_INST_OBJ(dmctx, parent_node, (void *)du_obj, inst) == DM_STOP)
			break;
	}
	return 0;
}

static int browseSoftwareModulesExecutionUnitInst(struct dmctx *dmctx, DMNODE *parent_node, void *prev_data, char *prev_instance)
{
	json_object *res = NULL, *du_obj = NULL, *arrobj = NULL;
	char *inst = NULL;
	int id = 0, eu = 0;

	dmubus_call("swmodules", "eu_list", UBUS_ARGS{0}, 0, &res);
	dmjson_foreach_obj_in_array(res, arrobj, du_obj, eu, 1, "execution_unit") {
		inst = handle_instance_without_section(dmctx, parent_node, ++id);
		if (DM_LINK_INST_OBJ(dmctx, parent_node, (void *)du_obj, inst) == DM_STOP)
			break;
	}
	return 0;
}

/*************************************************************
* COMMON FUNCTIONS
**************************************************************/
static int get_SoftwareModules_VendorConfigList(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	struct uci_section *s = NULL;
	dmasprintf(value, "");

	char *config = dmjson_get_value((json_object *)data, 1, "config");
	if (config == NULL || *config == '\0')
		return 0;

	uci_path_foreach_sections(bbfdm, DMMAP, "vcf", s) {
		char *name = NULL;

		dmuci_get_value_by_section_string(s, "name", &name);
		if (name && strcmp(name, config) == 0) {
			char *vcf_instance;
			dmuci_get_value_by_section_string(s, "vcf_instance", &vcf_instance);
			dmasprintf(value, "Device.DeviceInfo.VendorConfigFile.%s", vcf_instance);
			break;
		}
	}
	return 0;
}

/*************************************************************
* GET & SET PARAM
*************************************************************/
static int get_SoftwareModules_ExecEnvNumberOfEntries(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	json_object *res = NULL, *environment = NULL;
	size_t nbre_env = 0;

	dmubus_call("swmodules", "ee_list", UBUS_ARGS{0}, 0, &res);
	DM_ASSERT(res, dmasprintf(value, "0"));
	json_object_object_get_ex(res, "environment", &environment);
	nbre_env = (environment) ? json_object_array_length(environment) : 0;
	dmasprintf(value, "%d", nbre_env);
	return 0;
}

static int get_SoftwareModules_DeploymentUnitNumberOfEntries(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	json_object *res = NULL, *deployment_unit = NULL;
	size_t nbre_du = 0;

	dmubus_call("swmodules", "du_list", UBUS_ARGS{0}, 0, &res);
	DM_ASSERT(res, dmasprintf(value, "0"));
	json_object_object_get_ex(res, "deployment_unit", &deployment_unit);
	nbre_du = (deployment_unit) ? json_object_array_length(deployment_unit) : 0;
	dmasprintf(value, "%d", nbre_du);
	return 0;
}

static int get_SoftwareModules_ExecutionUnitNumberOfEntries(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	json_object *res = NULL, *execution_unit = NULL;
	size_t nbre_env = 0;

	dmubus_call("swmodules", "eu_list", UBUS_ARGS{0}, 0, &res);
	DM_ASSERT(res, dmasprintf(value, "0"));
	json_object_object_get_ex(res, "execution_unit", &execution_unit);
	nbre_env = (execution_unit) ? json_object_array_length(execution_unit) : 0;
	dmasprintf(value, "%d", nbre_env);
	return 0;
}

/*#Device.SoftwareModules.ExecEnv.{i}.Enable!UBUS:swmodules/environment//environment[i-1].autoboot*/
static int get_SoftwareModulesExecEnv_Enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmjson_get_value((json_object *)data, 1, "autoboot");
	return 0;
}

static int set_SoftwareModulesExecEnv_Enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	char *env_name = NULL;
	bool b;

	switch (action)	{
		case VALUECHECK:
			if (dm_validate_boolean(value))
				return FAULT_9007;
			break;
		case VALUESET:
			string_to_bool(value, &b);
			env_name = dmjson_get_value((json_object *)data, 1, "ee_name");
			char *op=(b ? "start":"stop");

			dmubus_call_set("swmodules", "ee_set_state", UBUS_ARGS{
				{"ee_name", env_name, String},
				{"state", op, String}},
				2);
			break;
	}
	return 0;
}

/*#Device.SoftwareModules.ExecEnv.{i}.Status!UBUS:swmodules/environment//environment[i-1].status*/
static int get_SoftwareModulesExecEnv_Status(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmjson_get_value((json_object *)data, 1, "status");
	return 0;
}

static int get_SoftwareModulesExecEnv_Reset(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmasprintf(value, "0");
	return 0;
}

static int set_SoftwareModulesExecEnv_Reset(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	char *env_name = NULL;
	bool b;

	switch (action)	{
		case VALUECHECK:
			if (dm_validate_boolean(value))
				return FAULT_9007;
			break;
		case VALUESET:
			string_to_bool(value, &b);
			env_name = dmjson_get_value((json_object *)data, 1, "ee_name");
			if (env_name && strcmp(env_name, "OpenWRT_Linux") == 0) {
				if (b) dmcmd_no_wait("/sbin/defaultreset", 0);
			}
			break;
	}
	return 0;
}

int get_SoftwareModulesExecEnv_Alias(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	struct uci_section *s = NULL;
	char *env_name = NULL;

	char *name = dmjson_get_value((json_object *)data, 1, "ee_name");
	uci_path_foreach_sections(bbfdm, "dmmap_sw_modules", "environment", s) {
		dmuci_get_value_by_section_string(s, "name", &env_name);
		if (name && env_name && strcmp(env_name, name) == 0) {
			dmuci_get_value_by_section_string(s, "alias", value);
			break;
		}
	}
	if ((*value)[0] == '\0')
		dmasprintf(value, "cpe-%s", instance);
	return 0;
}

int set_SoftwareModulesExecEnv_Alias(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	struct uci_section *s = NULL, *dmmap = NULL;
	char *name;
	bool found = false;

	switch (action)	{
		case VALUECHECK:
			if (dm_validate_string(value, -1, 64, NULL, NULL))
				return FAULT_9007;
			break;
		case VALUESET:
			name = dmjson_get_value((json_object *)data, 1, "ee_name");
			uci_path_foreach_option_eq(bbfdm, "dmmap_sw_modules", "environment", "name", name, s) {
				dmuci_set_value_by_section_bbfdm(s, "alias", value);
				found = true;
			}
			if (!found) {
				dmuci_add_section_bbfdm("dmmap_sw_modules", "environment", &dmmap);
				dmuci_set_value_by_section(dmmap, "name", name);
				dmuci_set_value_by_section(dmmap, "alias", value);
			}
			break;
	}
	return 0;
}

/*#Device.SoftwareModules.ExecEnv.{i}.Name!UBUS:swmodules/environment//environment[i-1].name*/
static int get_SoftwareModulesExecEnv_Name(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmjson_get_value((json_object *)data, 1, "ee_name");
	return 0;
}

/*#Device.SoftwareModules.ExecEnv.{i}.Type!UBUS:swmodules/environment//environment[i-1].type*/
static int get_SoftwareModulesExecEnv_Type(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmjson_get_value((json_object *)data, 1, "type");
	return 0;
}

/*#Device.SoftwareModules.ExecEnv.{i}.Vendor!UBUS:swmodules/environment//environment[i-1].vendor*/
static int get_SoftwareModulesExecEnv_Vendor(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmjson_get_value((json_object *)data, 1, "vendor");
	return 0;
}

/*#Device.SoftwareModules.ExecEnv.{i}.Version!UBUS:swmodules/environment//environment[i-1].version*/
static int get_SoftwareModulesExecEnv_Version(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmjson_get_value((json_object *)data, 1, "version");
	return 0;
}

static int get_SoftwareModulesExecEnv_ParentExecEnv(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *eid = dmjson_get_value((json_object *)data, 1, "parent_ee_ref");
	if (strcmp(eid, "0") == 0) {
		dmasprintf(value, "");
	} else {
		dmasprintf(value, "Device.SoftwareModules.ExecEnv.%s", eid);
	}

	return 0;
}

/*#Device.SoftwareModules.ExecEnv.{i}.AllocatedDiskSpace!UBUS:swmodules/environment//environment[i-1].allocateddiskspace*/
static int get_SoftwareModulesExecEnv_AllocatedDiskSpace(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmjson_get_value((json_object *)data, 1, "allocated_disk_space");
	return 0;
}

/*#Device.SoftwareModules.ExecEnv.{i}.AvailableDiskSpace!UBUS:swmodules/environment//environment[i-1].availablediskspace*/
static int get_SoftwareModulesExecEnv_AvailableDiskSpace(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmjson_get_value((json_object *)data, 1, "available_disk_space");
	return 0;
}

/*#Device.SoftwareModules.ExecEnv.{i}.AllocatedMemory!UBUS:swmodules/environment//environment[i-1].allocatedmemory*/
static int get_SoftwareModulesExecEnv_AllocatedMemory(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmjson_get_value((json_object *)data, 1, "allocated_memory");
	return 0;
}

/*#Device.SoftwareModules.ExecEnv.{i}.AvailableMemory!UBUS:swmodules/environment//environment[i-1].availablememory*/
static int get_SoftwareModulesExecEnv_AvailableMemory(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmjson_get_value((json_object *)data, 1, "available_memory");
	return 0;
}

int get_SoftwareModulesExecEnv_ActiveExecutionUnits(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	json_object *res = NULL, *du_obj = NULL, *arrobj = NULL;
	char eu_list[2048];
	unsigned pos = 0;
	int eu = 0;

	eu_list[0] = 0;
	char *curr_env = dmjson_get_value((json_object *)data, 1, "ee_name");
	dmubus_call("swmodules", "eu_list", UBUS_ARGS{0}, 0, &res);
	DM_ASSERT(res, dmasprintf(value, ""));
	dmjson_foreach_obj_in_array(res, arrobj, du_obj, eu, 1, "execution_unit") {
		char *environment = dmjson_get_value(du_obj, 1, "ee_name");

		if (strcmp(environment, curr_env) == 0)
			pos += snprintf(&eu_list[pos], sizeof(eu_list) - pos, "Device.SoftwareModules.ExecutionUnit.%d,", eu+1);
	}

	if (pos)
		eu_list[pos - 1] = 0;

	*value = dmstrdup(eu_list);
	return 0;
}

/*#Device.SoftwareModules.DeploymentUnit.{i}.UUID!UBUS:swmodules/du_list//deployment_unit[i-1].uuid*/
static int get_SoftwareModulesDeploymentUnit_UUID(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmjson_get_value((json_object *)data, 1, "uuid");
	return 0;
}

/*#Device.SoftwareModules.DeploymentUnit.{i}.DUID!UBUS:swmodules/du_list//deployment_unit[i-1].duid*/
static int get_SoftwareModulesDeploymentUnit_DUID(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmjson_get_value((json_object *)data, 1, "duid");
	return 0;
}

int get_SoftwareModulesDeploymentUnit_Alias(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	struct uci_section *s = NULL;
	char *du_name = NULL, *du_env = NULL;

	char *name = dmjson_get_value((json_object *)data, 1, "du_name");
	char *environment = dmjson_get_value((json_object *)data, 1, "ee_name");
	uci_path_foreach_sections(bbfdm, "dmmap_sw_modules", "deployment_unit", s) {
		dmuci_get_value_by_section_string(s, "name", &du_name);
		dmuci_get_value_by_section_string(s, "environment", &du_env);
		if (name && du_name && (strcmp(du_name, name) == 0) && (environment && du_env && strcmp(du_env, environment) == 0)) {
			dmuci_get_value_by_section_string(s, "alias", value);
			break;
		}
	}
	if ((*value)[0] == '\0')
		dmasprintf(value, "cpe-%s", instance);
	return 0;
}

int set_SoftwareModulesDeploymentUnit_Alias(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	struct uci_section *s = NULL, *dmmap = NULL;
	char *du_name, *du_env, *environment = NULL, *name = NULL;
	bool found = false;

	switch (action)	{
		case VALUECHECK:
			if (dm_validate_string(value, -1, 64, NULL, NULL))
				return FAULT_9007;
			break;
		case VALUESET:
			name = dmjson_get_value((json_object *)data, 1, "du_name");
			environment = dmjson_get_value((json_object *)data, 1, "ee_name");
			uci_path_foreach_sections(bbfdm, "dmmap_sw_modules", "deployment_unit", s) {
				dmuci_get_value_by_section_string(s, "name", &du_name);
				dmuci_get_value_by_section_string(s, "environment", &du_env);
				if (name && (strcmp(du_name, name) == 0) && (environment && strcmp(du_env, environment) == 0)) {
					dmuci_set_value_by_section_bbfdm(s, "alias", value);
					found = true;
					break;
				}
			}
			if (!found) {
				dmuci_add_section_bbfdm("dmmap_sw_modules", "deployment_unit", &dmmap);
				dmuci_set_value_by_section(dmmap, "name", name);
				dmuci_set_value_by_section(dmmap, "environment", environment);
				dmuci_set_value_by_section(dmmap, "alias", value);
			}
			break;
	}
	return 0;
}

/*#Device.SoftwareModules.DeploymentUnit.{i}.Name!UBUS:swmodules/du_list//deployment_unit[i-1].name*/
static int get_SoftwareModulesDeploymentUnit_Name(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmjson_get_value((json_object *)data, 1, "du_name");
	return 0;
}

static int get_SoftwareModulesDeploymentUnit_Status(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmasprintf(value, "Installed");
	return 0;
}

static int get_SoftwareModulesDeploymentUnit_Resolved(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmasprintf(value, "1");
	return 0;
}

/*#Device.SoftwareModules.DeploymentUnit.{i}.URL!UBUS:swmodules/du_list//deployment_unit[i-1].url*/
static int get_SoftwareModulesDeploymentUnit_URL(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmjson_get_value((json_object *)data, 1, "url");
	return 0;
}

/*#Device.SoftwareModules.DeploymentUnit.{i}.Description!UBUS:swmodules/du_list//deployment_unit[i-1].description*/
static int get_SoftwareModulesDeploymentUnit_Description(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmjson_get_value((json_object *)data, 1, "description");
	return 0;
}

/*#Device.SoftwareModules.DeploymentUnit.{i}.Vendor!UBUS:swmodules/du_list//deployment_unit[i-1].vendor*/
static int get_SoftwareModulesDeploymentUnit_Vendor(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmjson_get_value((json_object *)data, 1, "vendor");
	return 0;
}

/*#Device.SoftwareModules.DeploymentUnit.{i}.Version!UBUS:swmodules/du_list//deployment_unit[i-1].version*/
static int get_SoftwareModulesDeploymentUnit_Version(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmjson_get_value((json_object *)data, 1, "version");
	return 0;
}

/*#Device.SoftwareModules.DeploymentUnit.{i}.VendorConfigList!UBUS:swmodules/du_list//deployment_unit[i-1].config*/
int get_SoftwareModulesDeploymentUnit_VendorConfigList(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	return get_SoftwareModules_VendorConfigList(refparam, ctx, data, instance, value);
}

int get_SoftwareModulesDeploymentUnit_ExecutionUnitList(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	json_object *res = NULL, *du_obj = NULL, *arrobj = NULL;
	char *environment = NULL, *name = NULL;
	int eu = 0;

	char *curr_name = dmjson_get_value((json_object *)data, 1, "du_name");
	char *curr_environment = dmjson_get_value((json_object *)data, 1, "ee_name");

	dmubus_call("swmodules", "eu_list", UBUS_ARGS{0}, 0, &res);
	DM_ASSERT(res, dmasprintf(value, ""));
	dmjson_foreach_obj_in_array(res, arrobj, du_obj, eu, 1, "execution_unit") {
		name = dmjson_get_value(du_obj, 1, "eu_name");
		environment = dmjson_get_value(du_obj, 1, "ee_name");
		if ((name && curr_name && strcmp(name, curr_name) == 0) && (environment && curr_environment && strcmp(environment, curr_environment) == 0)) {
			dmasprintf(value, "Device.SoftwareModules.ExecutionUnit.%d", eu+1);
			break;
		}
	}
	return 0;
}

/*#Device.SoftwareModules.DeploymentUnit.{i}.ExecutionEnvRef!UBUS:swmodules/du_list//deployment_unit[i-1].environment*/
static int get_SoftwareModulesDeploymentUnit_ExecutionEnvRef(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *eid = dmjson_get_value((json_object *)data, 1, "eeid");

	dmasprintf(value, "Device.SoftwareModules.ExecEnv.%s", eid);
	return 0;
}

/*#Device.SoftwareModules.ExecutionUnit.{i}.EUID!UBUS:swmodules/eu_list//execution_unit[i-1].euid*/
static int get_SoftwareModulesExecutionUnit_EUID(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmjson_get_value((json_object *)data, 1, "euid");
	return 0;
}

int get_SoftwareModulesExecutionUnit_Alias(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	struct uci_section *s = NULL;
	char *eu_euid = NULL, *eu_env = NULL;

	char *euid = dmjson_get_value((json_object *)data, 1, "euid");
	char *environment = dmjson_get_value((json_object *)data, 1, "ee_name");
	uci_path_foreach_sections(bbfdm, "dmmap_sw_modules", "execution_unit", s) {
		dmuci_get_value_by_section_string(s, "euid", &eu_euid);
		dmuci_get_value_by_section_string(s, "environment", &eu_env);
		if ((euid && eu_euid && strcmp(eu_euid, euid) == 0) && (environment && eu_env && strcmp(eu_env, environment) == 0)) {
			dmuci_get_value_by_section_string(s, "alias", value);
			break;
		}
	}
	if ((*value)[0] == '\0')
		dmasprintf(value, "cpe-%s", instance);
	return 0;
}

int set_SoftwareModulesExecutionUnit_Alias(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	struct uci_section *s = NULL, *dmmap = NULL;
	char *eu_euid, *eu_env, *environment, *euid;
	bool found = false;

	switch (action)	{
		case VALUECHECK:
			if (dm_validate_string(value, -1, 64, NULL, NULL))
				return FAULT_9007;
			break;
		case VALUESET:
			euid = dmjson_get_value((json_object *)data, 1, "euid");
			environment = dmjson_get_value((json_object *)data, 1, "ee_name");
			uci_path_foreach_sections(bbfdm, "dmmap_sw_modules", "execution_unit", s) {
				dmuci_get_value_by_section_string(s, "euid", &eu_euid);
				dmuci_get_value_by_section_string(s, "environment", &eu_env);
				if ((strcmp(eu_euid, euid) == 0) && (strcmp(eu_env, environment) == 0)) {
					dmuci_set_value_by_section_bbfdm(s, "alias", value);
					found = true;
					break;
				}
			}
			if (!found) {
				dmuci_add_section_bbfdm("dmmap_sw_modules", "execution_unit", &dmmap);
				dmuci_set_value_by_section(dmmap, "euid", euid);
				dmuci_set_value_by_section(dmmap, "environment", environment);
				dmuci_set_value_by_section(dmmap, "alias", value);
			}
			break;
	}
	return 0;
}

/*#Device.SoftwareModules.ExecutionUnit.{i}.Name!UBUS:swmodules/eu_list//execution_unit[i-1].name*/
static int get_SoftwareModulesExecutionUnit_Name(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmjson_get_value((json_object *)data, 1, "eu_name");
	return 0;
}

/*#Device.SoftwareModules.ExecutionUnit.{i}.ExecEnvLabel!UBUS:swmodules/eu_list//execution_unit[i-1].euid*/
static int get_SoftwareModulesExecutionUnit_ExecEnvLabel(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmjson_get_value((json_object *)data, 1, "euid");
	return 0;
}

static int get_SoftwareModulesExecutionUnit_Status(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmjson_get_value((json_object *)data, 1, "state");
	return 0;
}

/*#Device.SoftwareModules.ExecutionUnit.{i}.Vendor!UBUS:swmodules/eu_list//execution_unit[i-1].vendor*/
static int get_SoftwareModulesExecutionUnit_Vendor(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmjson_get_value((json_object *)data, 1, "vendor");
	return 0;
}

/*#Device.SoftwareModules.ExecutionUnit.{i}.Version!UBUS:swmodules/eu_list//execution_unit[i-1].version*/
static int get_SoftwareModulesExecutionUnit_Version(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmjson_get_value((json_object *)data, 1, "version");
	return 0;
}

/*#Device.SoftwareModules.ExecutionUnit.{i}.Description!UBUS:swmodules/eu_list//execution_unit[i-1].description*/
static int get_SoftwareModulesExecutionUnit_Description(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmjson_get_value((json_object *)data, 1, "description");
	return 0;
}

/*#Device.SoftwareModules.ExecutionUnit.{i}.DiskSpaceInUse!UBUS:swmodules/eu_list//execution_unit[i-1].disk_space*/
static int get_SoftwareModulesExecutionUnit_DiskSpaceInUse(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmjson_get_value((json_object *)data, 1, "disk_space");
	return 0;
}

/*#Device.SoftwareModules.ExecutionUnit.{i}.MemoryInUse!UBUS:swmodules/eu_list//execution_unit[i-1].memory_space*/
static int get_SoftwareModulesExecutionUnit_MemoryInUse(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmjson_get_value((json_object *)data, 1, "memory_space");
	return 0;
}

int get_SoftwareModulesExecutionUnit_References(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	json_object *res = NULL, *du_obj = NULL, *arrobj = NULL;
	char *environment = NULL, *name = NULL;
	int du = 0;

	char *curr_name = dmjson_get_value((json_object *)data, 1, "eu_name");
	char *curr_environment = dmjson_get_value((json_object *)data, 1, "ee_name");

	dmubus_call("swmodules", "du_list", UBUS_ARGS{0}, 0, &res);
	DM_ASSERT(res, dmasprintf(value, ""));
	dmjson_foreach_obj_in_array(res, arrobj, du_obj, du, 1, "deployment_unit") {
		name = dmjson_get_value(du_obj, 1, "du_name");
		environment = dmjson_get_value(du_obj, 1, "ee_name");
		if ((name && curr_name && strcmp(name, curr_name) == 0) && (environment && curr_environment && strcmp(environment, curr_environment) == 0)) {
			dmasprintf(value, "Device.SoftwareModules.DeploymentUnit.%d", du+1);
			break;
		}
	}
	return 0;
}

int get_SoftwareModulesExecutionUnit_AssociatedProcessList(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	json_object *res = NULL, *processes_obj = NULL, *arrobj = NULL;
	char *pid = NULL;
	int process = 0;

	char *euid = dmjson_get_value((json_object *)data, 1, "euid");
	dmubus_call("router.system", "processes", UBUS_ARGS{0}, 0, &res);
	DM_ASSERT(res, dmasprintf(value, ""));
	dmjson_foreach_obj_in_array(res, arrobj, processes_obj, process, 1, "processes") {
		pid = dmjson_get_value(processes_obj, 1, "pid");
		if (pid && euid && strcmp(euid, pid) == 0) {
			dmasprintf(value, "Device.DeviceInfo.ProcessStatus.Process.%d", process+1);
			break;
		}
	}
	return 0;
}

/*#Device.SoftwareModules.ExecutionUnit.{i}.VendorConfigList!UBUS:swmodules/eu_list//execution_unit[i-1].config*/
int get_SoftwareModulesExecutionUnit_VendorConfigList(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	return get_SoftwareModules_VendorConfigList(refparam, ctx, data, instance, value);
}

/*#Device.SoftwareModules.ExecutionUnit.{i}.ExecutionEnvRef!UBUS:swmodules/eu_list//execution_unit[i-1].environment*/
static int get_SoftwareModulesExecutionUnit_ExecutionEnvRef(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *eid = dmjson_get_value((json_object *)data, 1, "eeid");

	dmasprintf(value, "Device.SoftwareModules.ExecEnv.%s", eid);
	return 0;
}

/*************************************************************
 * OPERATE COMMANDS
 *************************************************************/
static int operate_SoftwareModulesExecEnv_Reset(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	char *name = dmjson_get_value((json_object *)data, 1, "ee_name");
	if (strcmp(name, "OpenWRT_Linux") == 0 && dmcmd_no_wait("/sbin/defaultreset", 0) == 0)
			return CMD_SUCCESS;
	return CMD_FAIL;
}

static operation_args softwaremodules_installdu_args = {
	.in = (const char *[]) {
		"URL",
		"UUID",
		"Username",
		"Password",
		"ExecutionEnvRef",
		NULL
	}
};

static int get_operate_args_SoftwareModules_InstallDU(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = (char *)&softwaremodules_installdu_args;
	return 0;
}

static int operate_SoftwareModules_InstallDU(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	json_object *res = NULL;

	char *url = dmjson_get_value((json_object *)value, 1, "URL");
	if (url[0] == '\0')
		return CMD_INVALID_ARGUMENTS;

	char *uuid = dmjson_get_value((json_object *)value, 1, "UUID");
	char *username = dmjson_get_value((json_object *)value, 1, "Username");
	char *password = dmjson_get_value((json_object *)value, 1, "Password");
	char *environment = dmjson_get_value((json_object *)value, 1, "ExecutionEnvRef");
	char *eid_ptr = strrchr(environment, '.');
	if (eid_ptr)
		eid_ptr = eid_ptr + 1;

	dmubus_call("swmodules", "du_install", UBUS_ARGS{
			{"url", url, String},
			{"uuid", uuid, String},
			{"username", username, String},
			{"password", password, String},
			{"eeid", eid_ptr, Integer}},
			5,
			&res);

	char *status = dmjson_get_value(res, 1, "status");

	return (strcmp(status, "true") == 0) ? CMD_SUCCESS : CMD_FAIL;
}

static operation_args softwaremodulesdeploymentunit_update_args = {
	.in = (const char *[]) {
		"URL",
		"Username",
		"Password",
		NULL
	}
};

static int get_operate_args_SoftwareModulesDeploymentUnit_Update(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = (char *)&softwaremodulesdeploymentunit_update_args;
	return 0;
}

static int operate_SoftwareModulesDeploymentUnit_Update(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	json_object *res = NULL;

	char *url = dmjson_get_value((json_object *)value, 1, "URL");
	if (url[0] == '\0')
		return CMD_INVALID_ARGUMENTS;

	char *eid = dmjson_get_value((json_object *)data, 1, "eeid");
	char *username = dmjson_get_value((json_object *)value, 1, "Username");
	char *password = dmjson_get_value((json_object *)value, 1, "Password");

	char *du_uuid = dmjson_get_value((json_object *)data, 1, "uuid");
	if (!du_uuid)
		return CMD_FAIL;

	dmubus_call("swmodules", "du_update", UBUS_ARGS{
			{"uuid", du_uuid, String},
			{"url", url, String},
			{"eeid", eid, Integer},
			{"username", username, String},
			{"password", password, String}},
			5,
			&res);

	char *status = dmjson_get_value(res, 1, "status");

	return (strcmp(status, "true") == 0) ? CMD_SUCCESS : CMD_FAIL;
}

static int operate_SoftwareModulesDeploymentUnit_Uninstall(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	json_object *res = NULL;

	char *du_name = dmjson_get_value((json_object *)data, 1, "du_name");
	char *eid = dmjson_get_value((json_object *)data, 1, "eeid");

	dmubus_call("swmodules", "du_uninstall", UBUS_ARGS{
			{"du_name", du_name, String},
			{"eeid", eid, Integer}},
			2,
			&res);

	char *status = dmjson_get_value(res, 1, "status");

	return (strcmp(status, "true") == 0) ? CMD_SUCCESS : CMD_FAIL;
}

static operation_args softwaremodulesexecutionunit_setrequestedstate_args = {
	.in = (const char *[]) {
		"RequestedState",
		NULL
	}
};

static int get_operate_args_SoftwareModulesExecutionUnit_SetRequestedState(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = (char *)&softwaremodulesexecutionunit_setrequestedstate_args;
	return 0;
}

static int operate_SoftwareModulesExecutionUnit_SetRequestedState(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	char *req_state = dmjson_get_value((json_object *)value, 1, "RequestedState");
	if (req_state[0] == '\0') {
		return CMD_INVALID_ARGUMENTS;
	}

	char *eid = dmjson_get_value((json_object *)data, 1, "eeid");
	char *name = dmjson_get_value((json_object *)data, 1, "eu_name");
	char *status = dmjson_get_value((json_object *)data, 1, "state");

	if (!eid || !name || !status) {
		return CMD_FAIL;
	}

	if (strcmp(req_state, "Idle") == 0 && (strcmp(status, "Active") == 0 || strcmp(status, "Starting") == 0)) {
		/* state can be changed to Idle only when current state is either Active or Starting */
		dmubus_call_set("swmodules", "eu_set_state", UBUS_ARGS{
				{"eeid", eid, Integer},
				{"eu_name", name, String},
				{"state", "false", Boolean}},
				3);
	} else if (strcmp(req_state, "Active") == 0 && strcmp(status, "Idle") == 0) {
		/* state can be changed to Active only when current state is Idle */
		dmubus_call_set("swmodules", "eu_set_state", UBUS_ARGS{
				{"eeid", eid, Integer},
				{"eu_name", name, String},
				{"state", "true", Boolean}},
				3);
	} else {
		/* Ignore the request */
		return CMD_SUCCESS;
	}

	return CMD_SUCCESS;
}

/*************************************************************
 * EVENTS
 *************************************************************/
static event_args du_state_change_args = {
	.param = (const char *[]) {
		"UUID",
		"DeploymentUnitRef",
		"Version",
		"CurrentState",
		"Resolved",
		"ExecutionUnitRefList",
		"StartTime",
		"CompleteTime",
		"OperationPerformed",
		"Fault.FaultCode",
		"Fault.FaultString",
		NULL
	}
};

static int get_event_DUStateChagne(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = (char *)&du_state_change_args;
	return 0;
}

/**********************************************************************************************************************************
*                                            OBJ & PARAM DEFINITION
***********************************************************************************************************************************/
/* *** Device. *** */
DMOBJ tSWmodObj[] = {
/* OBJ, permission, addobj, delobj, checkdep, browseinstobj, nextdynamicobj, dynamicleaf, nextobj, leaf, linker, bbfdm_type, uniqueKeys*/
{"SoftwareModules", &DMREAD, NULL, NULL, NULL, NULL, NULL, NULL, tSoftwareModulesObj, tSoftwareModulesParams, NULL, BBFDM_BOTH},
{0}
};

/* *** Device.SoftwareModules. *** */
DMOBJ tSoftwareModulesObj[] = {
/* OBJ, permission, addobj, delobj, checkdep, browseinstobj, nextdynamicobj, dynamicleaf, nextobj, leaf, linker, bbfdm_type, uniqueKeys*/
{"ExecEnv", &DMREAD, NULL, NULL, NULL, browseSoftwareModulesExecEnvInst, NULL, NULL, NULL, tSoftwareModulesExecEnvParams, get_exec_env_linker, BBFDM_BOTH, LIST_KEY{"Alias", "Name", NULL}},
{"DeploymentUnit", &DMREAD, NULL, NULL, NULL, browseSoftwareModulesDeploymentUnitInst, NULL, NULL, NULL, tSoftwareModulesDeploymentUnitParams, get_du_linker, BBFDM_BOTH, LIST_KEY{"UUID", "Version", "ExecutionEnvRef", "Alias", NULL}},
{"ExecutionUnit", &DMREAD, NULL, NULL, NULL, browseSoftwareModulesExecutionUnitInst, NULL, NULL, NULL, tSoftwareModulesExecutionUnitParams, NULL, BBFDM_BOTH, LIST_KEY{"EUID", "Alias", NULL}},
{0}
};

DMLEAF tSoftwareModulesParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type*/
{"ExecEnvNumberOfEntries", &DMREAD, DMT_UNINT, get_SoftwareModules_ExecEnvNumberOfEntries, NULL, BBFDM_BOTH},
{"DeploymentUnitNumberOfEntries", &DMREAD, DMT_UNINT, get_SoftwareModules_DeploymentUnitNumberOfEntries, NULL, BBFDM_BOTH},
{"ExecutionUnitNumberOfEntries", &DMREAD, DMT_UNINT, get_SoftwareModules_ExecutionUnitNumberOfEntries, NULL, BBFDM_BOTH},
{"InstallDU()", &DMASYNC, DMT_COMMAND, get_operate_args_SoftwareModules_InstallDU, operate_SoftwareModules_InstallDU, BBFDM_USP},
{"DUStateChange!", &DMREAD, DMT_EVENT, get_event_DUStateChagne, NULL, BBFDM_USP},
{0}
};

/* *** Device.SoftwareModules.ExecEnv.{i}. *** */
DMLEAF tSoftwareModulesExecEnvParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type*/
{"Enable", &DMWRITE, DMT_BOOL, get_SoftwareModulesExecEnv_Enable, set_SoftwareModulesExecEnv_Enable, BBFDM_BOTH},
{"Status", &DMREAD, DMT_STRING, get_SoftwareModulesExecEnv_Status, NULL, BBFDM_BOTH},
{"Reset", &DMWRITE, DMT_BOOL, get_SoftwareModulesExecEnv_Reset, set_SoftwareModulesExecEnv_Reset, BBFDM_CWMP},
//{"Alias", &DMWRITE, DMT_STRING, get_SoftwareModulesExecEnv_Alias, set_SoftwareModulesExecEnv_Alias, BBFDM_BOTH},
{"Name", &DMREAD, DMT_STRING, get_SoftwareModulesExecEnv_Name, NULL, BBFDM_BOTH},
{"Type", &DMREAD, DMT_STRING, get_SoftwareModulesExecEnv_Type, NULL, BBFDM_BOTH},
//{"InitialRunLevel", &DMWRITE, DMT_UNINT, get_SoftwareModulesExecEnv_InitialRunLevel, set_SoftwareModulesExecEnv_InitialRunLevel, BBFDM_BOTH},
//{"RequestedRunLevel", &DMWRITE, DMT_INT, get_SoftwareModulesExecEnv_RequestedRunLevel, set_SoftwareModulesExecEnv_RequestedRunLevel, BBFDM_CWMP},
//{"CurrentRunLevel", &DMREAD, DMT_INT, get_SoftwareModulesExecEnv_CurrentRunLevel, NULL, BBFDM_BOTH},
//{"InitialExecutionUnitRunLevel", &DMWRITE, DMT_INT, get_SoftwareModulesExecEnv_InitialExecutionUnitRunLevel, set_SoftwareModulesExecEnv_InitialExecutionUnitRunLevel, BBFDM_BOTH},
{"Vendor", &DMREAD, DMT_STRING, get_SoftwareModulesExecEnv_Vendor, NULL, BBFDM_BOTH},
{"Version", &DMREAD, DMT_STRING, get_SoftwareModulesExecEnv_Version, NULL, BBFDM_BOTH},
{"ParentExecEnv", &DMREAD, DMT_STRING, get_SoftwareModulesExecEnv_ParentExecEnv, NULL, BBFDM_BOTH},
{"AllocatedDiskSpace", &DMREAD, DMT_INT, get_SoftwareModulesExecEnv_AllocatedDiskSpace, NULL, BBFDM_BOTH},
{"AvailableDiskSpace", &DMREAD, DMT_INT, get_SoftwareModulesExecEnv_AvailableDiskSpace, NULL, BBFDM_BOTH},
{"AllocatedMemory", &DMREAD, DMT_INT, get_SoftwareModulesExecEnv_AllocatedMemory, NULL, BBFDM_BOTH},
{"AvailableMemory", &DMREAD, DMT_INT, get_SoftwareModulesExecEnv_AvailableMemory, NULL, BBFDM_BOTH},
//{"ActiveExecutionUnits", &DMREAD, DMT_STRING, get_SoftwareModulesExecEnv_ActiveExecutionUnits, NULL, BBFDM_BOTH},
//{"ProcessorRefList", &DMREAD, DMT_STRING, get_SoftwareModulesExecEnv_ProcessorRefList, NULL, BBFDM_BOTH},
{"Reset()", &DMSYNC, DMT_COMMAND, NULL, operate_SoftwareModulesExecEnv_Reset, BBFDM_USP},
{0}
};

/* *** Device.SoftwareModules.DeploymentUnit.{i}. *** */
DMLEAF tSoftwareModulesDeploymentUnitParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type*/
{"UUID", &DMREAD, DMT_STRING, get_SoftwareModulesDeploymentUnit_UUID, NULL, BBFDM_BOTH},
{"DUID", &DMREAD, DMT_STRING, get_SoftwareModulesDeploymentUnit_DUID, NULL, BBFDM_BOTH},
//{"Alias", &DMWRITE, DMT_STRING, get_SoftwareModulesDeploymentUnit_Alias, set_SoftwareModulesDeploymentUnit_Alias, BBFDM_BOTH},
{"Name", &DMREAD, DMT_STRING, get_SoftwareModulesDeploymentUnit_Name, NULL, BBFDM_BOTH},
{"Status", &DMREAD, DMT_STRING, get_SoftwareModulesDeploymentUnit_Status, NULL, BBFDM_BOTH},
{"Resolved", &DMREAD, DMT_BOOL, get_SoftwareModulesDeploymentUnit_Resolved, NULL, BBFDM_BOTH},
{"URL", &DMREAD, DMT_STRING, get_SoftwareModulesDeploymentUnit_URL, NULL, BBFDM_BOTH},
{"Description", &DMREAD, DMT_STRING, get_SoftwareModulesDeploymentUnit_Description, NULL, BBFDM_BOTH},
{"Vendor", &DMREAD, DMT_STRING, get_SoftwareModulesDeploymentUnit_Vendor, NULL, BBFDM_BOTH},
{"Version", &DMREAD, DMT_STRING, get_SoftwareModulesDeploymentUnit_Version, NULL, BBFDM_BOTH},
//{"VendorLogList", &DMREAD, DMT_STRING, get_SoftwareModulesDeploymentUnit_VendorLogList, NULL, BBFDM_BOTH},
//{"VendorConfigList", &DMREAD, DMT_STRING, get_SoftwareModulesDeploymentUnit_VendorConfigList, NULL, BBFDM_BOTH},
//{"ExecutionUnitList", &DMREAD, DMT_STRING, get_SoftwareModulesDeploymentUnit_ExecutionUnitList, NULL, BBFDM_BOTH},
{"ExecutionEnvRef", &DMREAD, DMT_STRING, get_SoftwareModulesDeploymentUnit_ExecutionEnvRef, NULL, BBFDM_BOTH},
{"Update()", &DMASYNC, DMT_COMMAND, get_operate_args_SoftwareModulesDeploymentUnit_Update, operate_SoftwareModulesDeploymentUnit_Update, BBFDM_USP},
{"Uninstall()", &DMASYNC, DMT_COMMAND, NULL, operate_SoftwareModulesDeploymentUnit_Uninstall, BBFDM_USP},
{0}
};

DMLEAF tSoftwareModulesExecutionUnitParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type*/
{"EUID", &DMREAD, DMT_STRING, get_SoftwareModulesExecutionUnit_EUID, NULL, BBFDM_BOTH},
//{"Alias", &DMWRITE, DMT_STRING, get_SoftwareModulesExecutionUnit_Alias, set_SoftwareModulesExecutionUnit_Alias, BBFDM_BOTH},
{"Name", &DMREAD, DMT_STRING, get_SoftwareModulesExecutionUnit_Name, NULL, BBFDM_BOTH},
{"ExecEnvLabel", &DMREAD, DMT_STRING, get_SoftwareModulesExecutionUnit_ExecEnvLabel, NULL, BBFDM_BOTH},
{"Status", &DMREAD, DMT_STRING, get_SoftwareModulesExecutionUnit_Status, NULL, BBFDM_BOTH},
//{"RequestedState", &DMWRITE, DMT_STRING, get_SoftwareModulesExecutionUnit_RequestedState, set_SoftwareModulesExecutionUnit_RequestedState, BBFDM_CWMP},
//{"ExecutionFaultCode", &DMREAD, DMT_STRING, get_SoftwareModulesExecutionUnit_ExecutionFaultCode, NULL, BBFDM_BOTH},
//{"ExecutionFaultMessage", &DMREAD, DMT_STRING, get_SoftwareModulesExecutionUnit_ExecutionFaultMessage, NULL, BBFDM_BOTH},
//{"AutoStart", &DMWRITE, DMT_BOOL, get_SoftwareModulesExecutionUnit_AutoStart, set_SoftwareModulesExecutionUnit_AutoStart, BBFDM_BOTH},
//{"RunLevel", &DMWRITE, DMT_UNINT, get_SoftwareModulesExecutionUnit_RunLevel, set_SoftwareModulesExecutionUnit_RunLevel, BBFDM_BOTH},
{"Vendor", &DMREAD, DMT_STRING, get_SoftwareModulesExecutionUnit_Vendor, NULL, BBFDM_BOTH},
{"Version", &DMREAD, DMT_STRING, get_SoftwareModulesExecutionUnit_Version, NULL, BBFDM_BOTH},
{"Description", &DMREAD, DMT_STRING, get_SoftwareModulesExecutionUnit_Description, NULL, BBFDM_BOTH},
{"DiskSpaceInUse", &DMREAD, DMT_INT, get_SoftwareModulesExecutionUnit_DiskSpaceInUse, NULL, BBFDM_BOTH},
{"MemoryInUse", &DMREAD, DMT_INT, get_SoftwareModulesExecutionUnit_MemoryInUse, NULL, BBFDM_BOTH},
{"References", &DMREAD, DMT_STRING, get_SoftwareModulesExecutionUnit_References, NULL, BBFDM_BOTH},
//{"AssociatedProcessList", &DMREAD, DMT_STRING, get_SoftwareModulesExecutionUnit_AssociatedProcessList, NULL, BBFDM_BOTH},
//{"VendorLogList", &DMREAD, DMT_STRING, get_SoftwareModulesExecutionUnit_VendorLogList, NULL, BBFDM_BOTH},
//{"VendorConfigList", &DMREAD, DMT_STRING, get_SoftwareModulesExecutionUnit_VendorConfigList, NULL, BBFDM_BOTH},
//{"SupportedDataModelList", &DMREAD, DMT_STRING, get_SoftwareModulesExecutionUnit_SupportedDataModelList, NULL, BBFDM_CWMP},
{"ExecutionEnvRef", &DMREAD, DMT_STRING, get_SoftwareModulesExecutionUnit_ExecutionEnvRef, NULL, BBFDM_BOTH},
{"SetRequestedState()", &DMSYNC, DMT_COMMAND, get_operate_args_SoftwareModulesExecutionUnit_SetRequestedState, operate_SoftwareModulesExecutionUnit_SetRequestedState, BBFDM_USP},
{0}
};
