/*
 * swmod.h: SWMOD deamon
 *
 * Copyright (C) 2020 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: Amin Ben Ramdhane <amin.benramdhane@pivasoftware.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#ifndef SWMOD_H
#define SWMOD_H

#include <stdbool.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubus.h>

#include "swmod_common.h"

#define HOST_SYSTEM "OpenWRT_Linux"
#define MAX_ENV 8 //Max Environment
#define MAX_EU 256 //Max Execution Unit
#define UBUS_TIMEOUT 5
#define MAX_BUFFER_LEN (102400)
#define MAX_LEN_16 16
#define MAX_LEN_32 32
#define MAX_LEN_64 64
#define MAX_LEN_128 128
#define MAX_LEN_512 512
#define MAX_LEN_1024 1024

typedef enum {
	EE_TYPE_UNKNOWN,
	EE_TYPE_HOST,
	EE_TYPE_LXC,
	EE_TYPE_MAX
} EE_TYPE_t;

typedef struct {
	bool exists;
	int parent_eeid;
	int eeid;
	EE_TYPE_t ee_type;
	int pause;
	int autoboot;
	unsigned long allocated_disk_space; //AllocatedDiskSpace
	unsigned long available_disk_space; //AvailableDiskSpace
	unsigned long allocated_memory; //AllocatedMemory
	unsigned long available_memory; //AvailableMemory
	struct list_head eu_list;
	char status[MAX_LEN_32]; //Status
	char name[MAX_LEN_32]; //Name
	char type[MAX_LEN_64]; //Type
	char vendor[MAX_LEN_128]; //Vendor
	char version[MAX_LEN_16]; //Version
	char map_du_name[MAX_LEN_64]; // name of map_du file which contains the du specific info
} ExecEnv_t;

typedef struct ExecUnit {
	bool eu_exists;
	int euid; //EUID
	int disk_space; //DiskSpaceInUse
	int memory_space; //MemoryInUse
	char name[MAX_LEN_32]; // //Execution Unit Name
	char state[MAX_LEN_32]; // //Execution Unit Name
	char command[MAX_LEN_32]; //Execution Unit Command
	char vendor[MAX_LEN_128]; //Execution Unit Vendor
	char config[MAX_LEN_32]; //Config File
	char version[MAX_LEN_32]; //Version
	char description[MAX_LEN_1024]; //Description
	char environment[MAX_LEN_32]; //Environment Name
} ExecUnit;

typedef struct EuNode {
	ExecUnit eu;
	struct list_head list;
} EuNode;

extern ExecEnv_t g_environments[MAX_ENV];

void swmod_send_event_du_state_change(bool failed, time_t start, time_t end, char *ver, int action, char *err);
#endif //SWMOD_H
