/*
 * swmod_host.c: SWMOD deamon
 *
 * Copyright (C) 2020 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: Amin Ben Ramdhane <amin.benramdhane@pivasoftware.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include <stdio.h>
#include <stdbool.h>
#include <sys/sysinfo.h>
#include <sys/utsname.h>
#include <sys/statvfs.h>

#include "tools.h"
#include "swmod_uci.h"
#include "swmod_host.h"
#include "swmod.h"

void populate_host_system_environment(void)
{
	g_environments[0].exists = true;
	g_environments[0].pause = 0;
	g_environments[0].ee_type = EE_TYPE_HOST;
	g_environments[0].autoboot = 1;
	g_environments[0].parent_eeid = 0;
	g_environments[0].eeid = 1;
	swmod_strncpy(g_environments[0].map_du_name, SWMOD_MAP_DU, MAX_LEN_64);
	swmod_strncpy(g_environments[0].name, HOST_SYSTEM, MAX_LEN_32);
	swmod_strncpy(g_environments[0].status, "Up", MAX_LEN_32);

	struct utsname utsname;

	if (uname(&utsname) >= 0) {
		swmod_strncpy(g_environments[0].type, utsname.sysname, MAX_LEN_64);
		swmod_strncpy(g_environments[0].vendor, utsname.nodename, MAX_LEN_128);
		swmod_strncpy(g_environments[0].version, utsname.release, MAX_LEN_16);
	}

	struct sysinfo sinfo;

	if (sysinfo(&sinfo) == 0) {
		g_environments[0].allocated_memory = (sinfo.totalram / 1024);
		g_environments[0].available_memory = (sinfo.freeram / 1024);
	}

	struct statvfs dinfo;
	if (statvfs("/", &dinfo) == 0) {
		g_environments[0].allocated_disk_space = (dinfo.f_bsize * dinfo.f_blocks) / 1024;
		g_environments[0].available_disk_space = (dinfo.f_bsize * dinfo.f_bfree) / 1024;
	} else {
		g_environments[0].allocated_disk_space = 0;
		g_environments[0].available_disk_space = 0;
	}

	create_file(SWMOD_PATH"/map_du");
}

void service_get_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	if (req == NULL) {
		PRINT_DEBUG("UBUS Req not received");
		return;
	}

	ExecEnv_t *env = (ExecEnv_t *)req->priv;
	if (env == NULL) {
		PRINT_DEBUG("ExecEnv not given");
		return;
	}

	update_eu_from_blob(env, msg);
}

int swmod_set_host_service_state(char *service, bool state)
{
	return ubus_call_service_state(service, state);
}
